# Copyright 1999 The Regents of the University of California
# All Rights Reserved
# 
# Permission to use, copy, modify and distribute any part of the
# GeoPlot tool for educational, research and non-profit
# purposes, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph
# and the following paragraphs appear in all copies.
# 
# Those desiring to incorporate this into commercial products or use
# for commercial purposes should contact the Technology Transfer
# Office, University of California, San Diego, 9500 Gilman Drive, La
# Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
# 
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
# PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
# DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# 
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
# UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
# SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
# OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
# OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
# PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
# ANY PATENT, TRADEMARK OR OTHER RIGHTS.
# 
# The GeoPlot tool is developed by Ram Periakaruppan at the University 
# of California, San Diego under the Cooperative Association for Internet 
# Data Analysis (CAIDA) Program. Support for this effort is provided by 
# the CAIDA grant NCR-9711092, and by CAIDA members.
# 
# Report bugs and suggestions to geoplot@caida.org.

package GeoPlot;

use GPMod::Node;
use GPMod::Link;
use GPMod::Key;
use GPMod::Path;

use Carp;              # for carp() and croak()

@ISA =  qw();
use strict;

use vars qw($AUTOLOAD);

sub new 
{
    my $this = shift;
    my $class = ref($this) || $this;

    if ($#_ < 6)
    {
	 croak "Invalid number of parameters to GeoPlot\nGeoPlot(mapname, mapwidth, mapheight, top_lat, top_lon, bot_lat, bot_lon)";
    }
     
    my ($mapname, $mapwidth, $mapheight, $top_lat, $top_lon, $bot_lat, 
	$bot_lon) = @_;

    foreach my $var ($mapwidth, $mapheight, $top_lat, $top_lon, $bot_lat, 
		     $bot_lon)
    {
	if (!($var =~ /^[\d|\-|\.]+$/))
	{
	     croak($var, " is not a number\n");
	}
    }

    my $self = {
		'mapname' => undef,
		'mapwidth' => undef,
		'mapheight' => undef,
		'bgcolor' => undef,       
		'fgcolor' => undef,       
		'title' => undef,         
		'date' => undef,          
		'default_node_size' => undef,     
		'default_line_size' => undef,     
		'default_node_color' => undef,
		'default_line_color' => undef,
		'default_font_size' => undef,
		'refresh_interval' => undef,
		'hide_key' => undef,
		'top_lat' => undef,               
		'top_lon' => undef,               
		'bot_lat' => undef,               
		'bot_lon' => undef,      
		'nodes' => undef,
		'links' => undef,
		'paths' => undef,
		'color_key' => undef,
		'size_key' => undef,
     };

     bless $self, $class;

     $self->{'mapname'} = $mapname;
     $self->{'mapwidth'} = $mapwidth;
     $self->{'mapheight'} = $mapheight;
     $self->{'top_lat'} = $top_lat;
     $self->{'top_lon'} = $top_lon;
     $self->{'bot_lat'} = $bot_lat;
     $self->{'bot_lon'} = $bot_lon;
     $self->_initialize();

     return $self;
}

sub _initialize
{
    my $self = shift;
    $self->{'bgcolor'} = "white";
    $self->{'fgcolor'} = "black";
    $self->{'title'}   = "";
    $self->{'date'}    = "";
    $self->{'default_node_size'} = 6;
    $self->{'default_line_size'} = 1;
    $self->{'default_node_color'} = "red";
    $self->{'default_line_color'} = "green";
    $self->{'default_font_size'} = 10;
    $self->{'refresh_interval'} = 0;
    $self->{'hide_key'} = "false";
    $self->{'nodes'} = {};
    $self->{'links'} = {};
    $self->{'paths'} = {};
}


sub AUTOLOAD
{
    my $self = shift;
    my $type = ref($self) || croak "$self is not an object";
    my $name = $AUTOLOAD;
    $name =~ s/.*://;
    unless (exists $self->{$name} )
    {
	croak "Can't access $name field in object of class $type";
    }
	
    if (@_)
    {
	return $self->{$name} = shift;
    }
    else
    {
	return $self->{$name};
    }
}

sub add_key
{
    my $self = shift;

    if ($#_ < 6)
    {
	 croak "Invalid number of parameters to add_color_key";
    }

    my ($type, $x, $y, $width, $height, $title, $tfsize, @key_datasets) = @_;

    foreach my $tmp ($x, $y, $width, $height, $tfsize)
    {
	if (!($tmp =~ /^[\d|\-|\.]+$/))
	{
	    croak($tmp, " is not a valid number\n");
	}
    }

    # check for key datasets
    if (not @key_datasets) 
    {
        carp "No key ranges for ", $type, " key\n";
        return 0;
    }       

    my $key = new Key($x, $y, $width, $height, $title, $tfsize, @key_datasets);
    
    if ($key != 0)
    {
	if ($type eq "color")
	{
	    $self->{'color_key'} = $key;
	}
    
	if ($type eq "size")
	{
	    $self->{'size_key'} = $key;
	}
    }
    return $key;
}

sub add_node
{
    my $self = shift;
    my %node_attributes = ();
    
    if ($#_ < 2)
    {
	 croak "Invalid number of parameters to add_node";
    }

    my ($name, $lat, $lon, $node_attributes_ref) = @_;

    if (!($lat =~ /^[\d|\-|\.]+$/))
    {
	croak($lat, " is not a valid number\n");
    }
    if (!($lon =~ /^[\d|\-|\.]+$/))
    {
	croak($lon, " is not a valid number\n");
    }

    # check fourth arg for hash
    if ($node_attributes_ref) 
    {
	if (ref($node_attributes_ref) eq "HASH") 
	{
	    %node_attributes = %{$node_attributes_ref};
	}
	else
	{
	    carp "The attributes of the node ", $name, " must be a hash";
	    return 0;
	}
    }

    my $node = new Node($name,$lat,$lon,%node_attributes);
    ${$self->{'nodes'}}{$name} = $node;

    return $node;
}


sub add_link
{
    my $self = shift;
    my %link_attributes = ();
    
    if ($#_ < 1)
    {
	 croak "Invalid number of parameters to add_link";
    }

    my ($src, $dest, $link_attributes_ref) = @_;

    # check third arg for hash
    if ($link_attributes_ref) 
    {
	if (ref($link_attributes_ref) eq "HASH") 
	{
	    %link_attributes = %{$link_attributes_ref};
	}
	else
	{
	    carp "The attributes of the link ", $src, "-->", $dest, 
	         " must be a hash";
	    return 0;
	}
    }

    my $link = new Link($src, $dest, %link_attributes);
    push @{$self->{'links'}{"$src\0$dest"}}, $link;

    return $link;
}

sub add_path
{
    my $self = shift;
    my ($path_attributes_ref, $nodes_ref) = @_;

    # check first arg for hash
    if ($path_attributes_ref && ref($path_attributes_ref) ne "HASH") 
    {
	carp "The attributes of the path must be a hash";
	return 0;
    }
 
    if (ref($nodes_ref) ne "ARRAY") 
    { 
	carp "The nodes that make up the path must be defined as an array";
	return 0;
    }   

    my $path = new Path($path_attributes_ref, $nodes_ref);

    ${$self->{'paths'}}{$path->p_nodes} = $path;
    return $path;
}

sub aggregate_paths
{
    my $self = shift;
    my ($type, $path_attributes_ref, $paths_ref) = @_;
    my ($src, $dest, $path, $tmpstr, $key) = ("","","","","");
    my (%links, %path_attributes);
    my $tmplink;

    # check first arg for hash
    if ($path_attributes_ref) 
    {
	if (ref($path_attributes_ref) eq "HASH") 
	{
	    %path_attributes = %{$path_attributes_ref};
	}
	else
	{
	    carp "The attributes of the aggregate path must be a hash";
	    return 0;
	}
    }
 
    if (ref($paths_ref) ne "ARRAY") 
    { 
	carp "The paths must be defined as an array";
	return 0;
    }   

    my $array_len = @{$paths_ref};

    for (my $i=0; $i < $array_len; $i++)
    {
	$path = ${$paths_ref}[$i];
    
        my $path_len = $#{$path->{'p_nodes'}};
        for (my $j=0; $j < $path_len; $j++)
	{
	    $src = $path->{'p_nodes'}[$j];
	    $dest = $path->{'p_nodes'}[$j+1];

	    if (ref($src) eq "Node")
	    {
		$src = $src->name;
	    }

	    if (ref($dest) eq "Node")
	    {
		$dest = $dest->name;
	    }
		
	    $tmpstr = $src."\0".$dest;
	    
	    if ($links{$tmpstr})
	    {
		$tmplink = $links{$tmpstr};
		
		if ($type eq "color")
		{
		    $path_attributes{"vcolor"} = $tmplink->value_color 
			                                  + $path->value_color;
		}
		elsif ($type eq "size")
		{
		    $path_attributes{"vsize"} = $tmplink->value_size 
			                                   + $path->value_size;
		}
		elsif ($type eq "colorsize")
		{
		    $path_attributes{"vcolor"} = $tmplink->value_color 
			                                 + $path->value_color;
		    $path_attributes{"vsize"} = $tmplink->value_size 
			                                 + $path->value_size;
		}

		my $link = new Link($src, $dest, %path_attributes);
		$links{$tmpstr} = $link;
		$tmplink = undef;
	    }
	    else
	    {
		my %tmp_attributes = $path->get_attributes;
		#print "direc = ", $tmp_attributes{"direc"}, "\n";
		my $link = new Link($src, $dest, %tmp_attributes);
		$links{$tmpstr} = $link;
	    }
	}
    
        delete(${$self->{'paths'}}{$path->p_nodes});
    }

    foreach $key (keys %links)
    {
	push @{$self->{'links'}{$key}}, $links{$key};
    }
}


sub delete_node
{
    my $self = shift;
    my ($node) = @_;

    if (ref($node) eq "Node")
    {
	delete(${$self->{'nodes'}}{$node->name});
    }
    else
    {
	delete(${$self->{'nodes'}}{$node});
    }
}


sub delete_key
{
    my $self = shift;
    my ($type) = @_;

    if ($type eq "color")
    {
	$self->{'color_key'} = undef;
    }
    elsif ($type eq "size")
    {
	$self->{'size_key'} = undef;
    }
}


sub delete_link
{
    my $self = shift;
    my ($src, $dest) = @_;
    
    if (ref($src) eq "Link")
    {
	my $tmpstr = $src->src . "\0" . $src->dest;
	delete($self->{'links'}{$tmpstr});
    }
    else
    {
	my $tmpstr = $src . "\0" . $dest;
	delete($self->{'links'}{$tmpstr});
    } 
}

sub delete_path
{
    my $self = shift;
    my ($path) = @_;
    
    if (ref($path) eq "Path")
    {
	delete($self->{'paths'}{$path->p_nodes});
    }
}


sub generate_html
{
	my $self = shift;
	my $file_ref = shift || \*STDOUT;
	#redirect everthing to the file that is given as a parameter 	
	$self->print_header($file_ref);
	$self->print_body($file_ref);
	$self->print_java_trailer($file_ref);
	$self->print_trailer($file_ref);
}

sub print_body {
    my $self = shift;
    my $file_ref = shift || \*STDOUT;
    my ($node_name, $link_name, $link, $path_name);
    #my $mapname = $self->{'mapname'};

    print $file_ref <<HTML_BODY;
        <applet code="GeoPlot.class" archive="GeoPlot.jar" width=$self->{'mapwidth'} height=$self->{'mapheight'}>
	   <param name=mapname value="$self->{'mapname'}">
	   <param name=bgcolor value="$self->{'bgcolor'}">
	   <param name=fgcolor value="$self->{'fgcolor'}">
	   <param name=title value="$self->{'title'}">
	   <param name=date value="$self->{'date'}">
	   <param name=default_node_size value=$self->{'default_node_size'}>
	   <param name=default_line_size value=$self->{'default_line_size'}>
	   <param name=default_node_color value="$self->{'default_node_color'}">
	   <param name=default_line_color value="$self->{'default_line_color'}">
	   <param name=default_font_size value=$self->{'default_font_size'}>
	   <param name=refresh_interval value=$self->{'refresh_interval'}>
	   <param name=hide_key value="$self->{'hide_key'}">
	   <param name=top_lat value=$self->{'top_lat'}>
	   <param name=top_lon value=$self->{'top_lon'}>
	   <param name=bot_lat value=$self->{'bot_lat'}>
	   <param name=bot_lon value=$self->{'bot_lon'}>
HTML_BODY

    if ($self->{'color_key'})
    {
        print $file_ref "\t   <param name=color_key value=\"", "\n";
        print $file_ref $self->{'color_key'}->print;
    }

    if ($self->{'size_key'})
    {
        print $file_ref "\t   <param name=size_key value=\"", "\n";
        print $file_ref $self->{'size_key'}->print;
    }

    print $file_ref "\t   <param name=data value=\"", "\n";


    foreach $node_name (sort keys %{$self->{'nodes'}})
    {
       print $file_ref "\t\t", ${$self->{'nodes'}}{$node_name}->print;
    }

    foreach $link_name (sort keys %{$self->{'links'}})
    {
        foreach $link (@{$self->{'links'}{$link_name}})
        {
            print $file_ref "\t\t", $link->print;
        }
    }

    foreach $path_name (keys %{$self->{'paths'}})
    {
       print $file_ref ${$self->{'paths'}}{$path_name}->print;
    }
}



sub print_header {
    my $self = shift;
    my $file_ref = shift || \*STDOUT;

    print $file_ref <<"HTML_HEADER";
<html>
<head>
      <title> GeoPlot </title>
</head>
<body bgcolor="ffffff">
<h1><center>GeoPlot</center></h1>
<hr>
<center>
HTML_HEADER
}

sub print_java_trailer {
    my $self = shift;
    my $file_ref = shift || \*STDOUT;

    print $file_ref <<"HTML_JAVA_TRAILER";
           ">
        Your Browser does not support Java!!!
        </applet>
</center>
<hr>
HTML_JAVA_TRAILER
}

sub print_trailer {
    my $self = shift;
    my $file_ref = shift || \*STDOUT;

    print $file_ref <<"HTML_TRAILER";
</body>
</html>
HTML_TRAILER
}

1;
